#!/usr/bin/env python3
"""
Copyright (C) 2022-2025 Space Science and Engineering Center (SSEC), University of Wisconsin-Madison.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

This file is part of the CSPP Geo Gridded GLM software package. CSPP Geo
Gridded GLM takes GOES GLM Level 2 LCFA files and grids them to the ABI
fixed grid. It does this using the open source glmtools python package by
Eric Bruning. glmtools can be found in the runtime directory of this package
or at https://github.com/deeplycloudy/glmtools.
"""

"""
Functions for creating a statistics file from a stack of Gridded GLM tiles.

Statistics files document potentially missing ("punctured") tiles and are 
required by AWIPS for it to perform multi-minute aggregations.
"""

import os
from datetime import datetime
import re
from typing import Iterable
import netCDF4


def create_statistics_filename(tile_filename: str) -> str:
    # ex: OR_GLM-L2-GLMF-M6_G16_statistics_e20220907012700.nc
    """
    It appears the quick way to produce this convention is to use a donor tile file and:
        * replace the T?? with "statistics"
        * add an e before the end time
    """
    statistics_filename = re.sub(r"_T[0-9]{2}_", "_statistics_", tile_filename)
    statistics_filename = re.sub(r"_([0-9]{14}).nc", r"_e\g<1>.nc", statistics_filename)
    return statistics_filename


def get_tile_number_from_tile_file(tile_filename: str) -> int:
    match = re.search(r"_T([0-9]{2})_", tile_filename)
    if not match:
        raise RuntimeError
    # match.group(0) is the entire matched string
    return int(match.group(1))


def get_tile_numbers(tile_files: list) -> list:
    return [get_tile_number_from_tile_file(tf) for tf in tile_files]


def get_minute_from_tile_file(tile_filename: str) -> datetime:
    match = re.search(r"_([0-9]{14}).nc$", tile_filename)
    if not match:
        raise RuntimeError
    # match.group(0) is the entire matched string
    return datetime.strptime(match.group(1), '%Y%m%d%H%M%S')


"""
variables:

dimensions:
    produced = 19 ;
    defined = 62 ;
variables:
        short produced(produced) ;
                produced:_FillValue = 0s ;
                produced:standard_name = "Tiles Produced" ;
                produced:long_name = "Tiles Produced" ;
                produced:_Unsigned = "true" ;
                produced:_Storage = "chunked" ;
                produced:_ChunkSizes = 19 ;
                produced:_DeflateLevel = 1 ;
                produced:_Shuffle = "true" ;
                produced:_Endianness = "little" ;
        short defined(defined) ;
                defined:_FillValue = 0s ;
                defined:standard_name = "Tiles Defined" ;
                defined:long_name = "Tiles Defined" ;
                defined:_Unsigned = "true" ;
                defined:_Storage = "chunked" ;
                defined:_ChunkSizes = 62 ;
                defined:_DeflateLevel = 1 ;
                defined:_Shuffle = "true" ;
                defined:_Endianness = "little" ;
"""

def create_statistics_file(tile_files: list,
                           minute: datetime,
                           ) -> str:
    good_tiles = [tile for tile in tile_files if get_minute_from_tile_file(tile) == minute]

    assert(len(good_tiles) > 0)

    donorTile = good_tiles[0]
    statistics_filename = create_statistics_filename(donorTile)
    assert(not os.path.exists(statistics_filename))
    
    tileNumbers = get_tile_numbers(good_tiles)

    ds = netCDF4.Dataset(statistics_filename, 'w')

    dimDefined = ds.createDimension('defined', size=62)
    dimProduced = ds.createDimension('produced', size=len(good_tiles))

    varDefined = ds.createVariable('defined', 
                                   datatype='i2',
                                   dimensions=(dimDefined),
                                   zlib=True, complevel=1,
                                   endian='little',
                                   fill_value=0,
                                   )
    varDefined.standard_name = "Tiles Defined"
    varDefined.long_name = "Tiles Defined"
    varDefined._Unsigned = "true"

    #   short defined
    #      defined = _, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 
    #    19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 
    #    37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 
    #    55, 56, 57, 58, 59, 60, 61 ;
    varDefined[:] = range(62)

    varProduced = ds.createVariable('produced',
                                    datatype='i2', 
                                    dimensions=(dimProduced),
                                    zlib=True, complevel=1,
                                    endian='little',
                                    fill_value=0,
                                    )
    varProduced.standard_name = "Tiles Produced"
    varProduced.long_name = "Tiles Produced"
    varProduced._Unsigned = "true"
    
    #   short produced 
    #     produced = 5, 7, 11, 12, 14, 17, 18, 19, 22, 23, 24, 25, 27, 31, 32, 35, 44, 49, 50 ;
    varProduced[:] = tileNumbers

    """
    attributes that should be copied directly from the donor tile file:
        :orbital_slot = "GOES-East" ;
        :platform_ID = "G16" ;
        :production_site = "GCP" ;
        :scene_id = "Full Disk" ;
        :spatial_resolution = "2km at nadir" ;
        :time_coverage_end = "2022-09-07T01:27:00Z" ;
        :time_coverage_start = "2022-09-07T01:26:00Z" ;
    """
    donor_ds = netCDF4.Dataset(donorTile, 'r')
    for attrName in ['orbital_slot',
                     'platform_ID',
                     'production_site',
                     'scene_id',
                     'spatial_resolution',
                     'time_coverage_end',
                     'time_coverage_start',
                     ]:
        setattr(ds, attrName, getattr(donor_ds, attrName))
    donor_ds.close()

    """
    attributes that need to be generated:
        :dataset_name = "OR_GLM-L2-GLMF-M6_G16_statistics_e20220907012700.nc" ;
        :title = "GLM Punctured Gridded Tiles Production Record" ;
    """
    ds.dataset_name = statistics_filename
    ds.title = "GLM Punctured Gridded Tiles Production Record"

    """
    attributes we're skipping unless we learn they're necessary:
            :production_data_source = "Realtime" ;
            :production_environment = "OE" ;
            :project = "GOES" ;
            :timeline_id = "ABI Mode 6" ;
    """

    ds.close()
    return statistics_filename
